/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tstore.h"

#define rdtscll(val) __asm__ __volatile__ ("rdtsc" : "=A" (val))

uint64_t hash_size;
uint64_t buckets;
uint64_t pri_buckets;
uint64_t sec_buckets;

#ifdef rdtscll
long long start_time;
long long hash_usage;
#endif

void init_hash()
{
#ifdef rdtscll
  rdtscll(start_time);
  hash_usage=0;
#endif
  if (buckets==0) buckets=3;
  if (buckets==1) {
    sec_buckets=0;
  } else {
    sec_buckets=1+((buckets-2)>>8);
    if (sec_buckets>1) sec_buckets=nextprime(sec_buckets);
  }
  pri_buckets=buckets-sec_buckets;
  hash_size=prevprime(pri_buckets<<16);
  fprintf(stderr,"%lld buckets (%lld primary %lld secondary)\n",
	  buckets,pri_buckets,sec_buckets);
}

uint64_t hash_sector(uint8_t *data)
{
  uint64_t r=0;
  int i;
#ifdef rdtscll
  long long t1,t2;
  rdtscll(t1);
#endif
  for (i=0;i<512;++i)
    r=((r<<8)+data[i])%hash_size;
#ifdef rdtscll
  rdtscll(t2);
  hash_usage += (t2-t1);
#endif
  return r;
}

void hash_advice(int overflows, int maxused, int maxsecused)
{
#ifdef rdtscll
  long long total_time;
  rdtscll(total_time);
  total_time -= start_time;

  printf("%lld%% time used for hashing\n",(hash_usage*100)/total_time);
#endif
  printf("%d primary bucket overflow%s\n",overflows,(overflows==1)?"":"s");
  printf("Largest primary hash-bucket %d of 73 slots used\n",maxused);
  printf("Largest secondary hash-bucket %d of 73 slots used\n",maxsecused);
  if (maxsecused>2)
    printf("Rehashing suggested\n");
}

int put_hash(uint8_t *bucket, uint64_t bnum, uint16_t small_hash, int *max)
{
  int r=0;
  int used=bucket[0];

  if (used<73) {
    /* fits */
    r=1;
    PUT_BNUM(bucket+(1+7*used),bnum);
    PUT_HASH(bucket+(6+7*used),small_hash);
  }

  if (used<255) ++used;
  if (used>*max) *max=used;
  bucket[0]=used;

  return r;
}
